/*
 * RiskAversionModel.java
 *
 * Created on 16. bezen 2007, 19:52
 *
 * This model for Repast is a simple simulation
 * of evolution of aversion to risk and uncertainty
 * 
 * 
 */

package riskaversion;

/**
 *
 * @author Jan Burian, email: honzaburian at seznam.vse.cz 
 */
import uchicago.src.sim.engine.SimpleModel;
import uchicago.src.sim.analysis.DataSource;
import uchicago.src.sim.analysis.Histogram;
import uchicago.src.sim.analysis.OpenSequenceGraph;
//import uchicago.src.sim.analysis.Plot;
//import uchicago.src.sim.analysis.plot.RepastPlotBox;
import uchicago.src.sim.analysis.plot.RepastPlot;
import uchicago.src.sim.analysis.OpenStats;
import uchicago.src.sim.analysis.Sequence;
import uchicago.src.sim.analysis.BinDataSource;

import java.util.ArrayList;
import java.awt.Color;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.FileNotFoundException;

public class RiskAversionModel extends SimpleModel { 
  
  private ArrayList birthList = new ArrayList();
  private ArrayList deathList = new ArrayList();
  private ArrayList winnerList = new ArrayList(); //is used only by fixed evolution
     
  private int maxAgentAge = 7;
  private int actualAgents;
//Actual number of agents
  private int numberOfAgents =250;
//Initial number of agents
  private double deathProb = 0.1;
//Probability of death if playing the game
  private double mutationRange =0.1;
//range of random value added to riskAversion of newborn agent
  private double winnerAdvantage =1.0;
//The coeficient for basReproductionProb after a succesfull game
  private boolean autoReproductionProb =false; 
//If True then baseReproductionProb will be set to 1/numberOfAgents
  private boolean fixedPopulation =true; 
//If True then the number of agents will be kept fixed
  public boolean sexualReproduction = true;   
  private double baseReproductionProb = 0.03;
  private int maxIterations = 3000;
  
  private double meanAversion [];
  private double globalMeanAversion;
  private double globalMinAversion;
  private double globalMaxAversion;
   
  private boolean printOutput = false; 
  private Histogram aversionHistogram;
  private MyPlot aversionPlot;
  private RepastPlot rPlot;

  private int populations=20;
  private ArrayList agentPopulation[];
 
  public RiskAversionModel () {
   name = "Risk Aversion Evolution";
   params = new String[]{"MaxAgentAge", "NumberOfAgents", "DeathProb",
   "MutationRange", "WinnerAdvantage", "AutoReproductionProb",
   "BaseReproductionProb", "FixedPopulation","PrintOutput","SexualReproduction",
   "Populations","MaxIterations"};
   }

  public RiskAversionAgent getAgent(int i) {
    return ((RiskAversionAgent) agentList.get(i));
  }
  
  public int getAgentNumber() {
    return (agentList.size());
  }
  
  public double getMeanAversion() {
    int size = agentList.size();
    double sumAversion = 0;
    for (int i = 0; i < size; i++) {
        RiskAversionAgent a = (RiskAversionAgent)agentList.get(i);
        sumAversion = sumAversion + a.rAversion;
    }
    return (sumAversion/ (double) size);
  }
  
  public void updateGlobalMeanMinMaxAversion() {
    double min = 1;
    double max =0; 
    double sumAversion = 0;
    double aversion = 0;
    for (int j = 0; j < populations; j++) {
      aversion = meanAversion[j];
      sumAversion = sumAversion + aversion;
      if (min>aversion) min=aversion;
      if (max<aversion) max=aversion;
     }
    globalMeanAversion = sumAversion / (double) populations;
    globalMinAversion = min;
    globalMaxAversion = max;
   // System.out.println("GlobalAversion:"+globalMeanAversion+"min" +
     //       globalMinAversion+"max"+globalMaxAversion);
  }
  
  public void removeAgent(RiskAversionAgent a) {
    deathList.add(a);
  }

  public void addAgent(RiskAversionAgent a) {
    birthList.add(a);
  }
  
  public void addWinner(RiskAversionAgent a) {
    winnerList.add(a);
  }
  
  public void unboundEvolution ()  {
    double rProb;
    int size = agentList.size();
   
    if (autoReproductionProb)
     baseReproductionProb = 1/(double)size;
    
    for (int i = 0; i < size; i++) {
        RiskAversionAgent a = (RiskAversionAgent)agentList.get(i);

        if (a.winner == 1) {
          rProb = baseReproductionProb * (1d+winnerAdvantage);
          a.winner=0;
        }
        else 
          rProb = baseReproductionProb;

        //System.out.println("rProb:"+rProb);

        if (getNextDoubleFromTo(0d,1d)<rProb)
          a.reproduction(mutationRange);

       // a.checkAge(maxAgentAge);
      }
  }

  public void fixedEvolution () {
   /* The advantage of winners is implemented by adding the fraction of winners to all agents
    multiplied by winnerAdvantage to 1. Then we generate a random number between 0 and this
    number. The random number multiplied by the size is the index of agent to reproduce. 
    If there will be no winners the reproduction probability
    is just 1/size. If there are n winners their reproduction probability is
    (1+winneradvantage)/size.   
    */    
    int size = agentList.size();
    int winnerNumber = winnerList.size();
    double winnerIndex = ((double)winnerNumber / (double)size) * winnerAdvantage ;
    int births = numberOfAgents - size;
    double reprIndex;
    RiskAversionAgent a;
            
    while (births>0) {
      reprIndex = getNextDoubleFromTo(0d,1d + winnerIndex );
      if (reprIndex<1) {
        a = (RiskAversionAgent)agentList.get((int) (reprIndex*(double)size));
   //change made for non fatal failures     
        if (a.winner < 2) { //this means the agent is actually not a looser
          if (printOutput) System.out.println(" Winners:"+ winnerNumber+
                " reprindex:"+reprIndex+" aNumber:"+(int)(reprIndex*(double)size)+" Aversion:"+a.rAversion+" Winner:"+a.winner); 
          a.reproduction(mutationRange);
          births --;
        } 
      }
      else {
        a = (RiskAversionAgent) winnerList.get((int)(((reprIndex-1d)/winnerIndex)*(double)winnerNumber));
        if (printOutput)    System.out.println(" Winners:"+ winnerNumber+
              " reprindex:"+reprIndex+" aNumber:"+(int)(((reprIndex-1d)/winnerIndex)*(double)winnerNumber)+
              " Aversion:"+a.rAversion+" Winner:"+a.winner); 
        a.reproduction(mutationRange);
        births --;
      }
    }
    
    winnerList.clear();
    
   /* for (int i = 0; i < size; i++) {
      a = (RiskAversionAgent)agentList.get(i);
      a.winner=false;
     // a.checkAge(maxAgentAge);
    } */
  }

  public void evolution ()  {
    if (fixedPopulation)
      fixedEvolution ();
    else
      unboundEvolution();
  }
  
  public void setup() {
    super.setup();
   
    generateNewSeed();
  }

  public void buildDisplay () {
    
    if (aversionHistogram != null) aversionHistogram.dispose();
    aversionHistogram = new Histogram("Aversion Distribution in 3000th iteration, Max age="+maxAgentAge, 20, 0d, 1.000001d, this);
    aversionHistogram.setYRange(0, 100.0);
    aversionHistogram.setStatsVisible(true);
    BinDataSource source = new BinDataSource()  {
      public double getBinValue(Object o) {
        RiskAversionAgent agent = (RiskAversionAgent)o;
        return agent.rAversion;
      }
    };
    aversionHistogram.createHistogramItem("Aversion", agentPopulation[0], source,-1,2);
    aversionHistogram.setSize(640,400);
    aversionHistogram.setSnapshotFileName("aversion_histogram_deathprob"+deathProb+
             "_advantage"+winnerAdvantage+
             "_age"+maxAgentAge+"_tick");
    aversionHistogram.display();
    
     if (aversionPlot != null) aversionPlot.dispose();
     aversionPlot = new MyPlot("Mean Aversion - Agent age="+maxAgentAge+
             ", Death probability=" + deathProb +
             ", Winner advantage="+ winnerAdvantage, this);
     aversionPlot.setConnected(true); 
     aversionPlot.setYRange(0d, 1d);
     aversionPlot.setXRange(0d,(double)maxIterations);
     aversionPlot.setAxisTitles("Time", "Average Mean");
     aversionPlot.addLegend(0, "AversionMean", Color.black);
     aversionPlot.setMarksStyle("dots",0);
     aversionPlot.addLegend(1, "AversionMean", Color.gray);
     aversionPlot.setMarksStyle("none",1);
     aversionPlot.setSnapshotFileName("aversion_graph_deathprob"+deathProb+
             "_advantage"+winnerAdvantage+
             "_age"+maxAgentAge+"_tick");
     aversionPlot.setSize(640,400);
     aversionPlot.display();
     
    //PrintData ();
    
  }

  public void PrintData () {
     
    winnerAdvantage=32;
    deathProb=0.3;
    MyPlot finalPlot = new MyPlot("Aversion on Age dependency , Winner Advantage="
            +winnerAdvantage+", Death Prob.="+deathProb, this);
     finalPlot.setConnected(true); 
     finalPlot.setYRange(0d, 1d);
     finalPlot.setXRange(0d,30);
     finalPlot.setAxisTitles("Max age of Agent", "Aversion Mean");
     finalPlot.addLegend(0, "AversionMean", Color.black);
     finalPlot.setMarksStyle("dots",0);
     finalPlot.addLegend(1, "AversionMean", Color.gray);
     finalPlot.setMarksStyle("none",1);
     finalPlot.setSnapshotFileName("aversionxage_deathprob"+deathProb+
             "_advantage"+winnerAdvantage+"_tick");
     finalPlot.setSize(800,600);
     finalPlot.display();
    
    
    finalPlot.updateGraph();
    finalPlot.takeSnapshot();
    stop();
  }
  
  public void buildModel() {
      
    birthList.clear(); 
    deathList.clear();
    winnerList.clear();  
    
    agentPopulation = new ArrayList [populations];
    
    System.gc();
    
    for(int j=0; j<populations; j++) {
       agentPopulation[j] = new ArrayList();
       for (int i = 0; i<numberOfAgents; i++) {
          RiskAversionAgent a = 
                  new RiskAversionAgent(getNextDoubleFromTo(0d,1d),
                                        getNextIntFromTo(0,maxAgentAge),this);
          agentPopulation[j].add(a);
         }
     }   
   
    meanAversion = new double [populations] ;
    
   buildDisplay ();
  }

  public String  ReadOldData (String fileName ) {
     /*Open file using BufferedReader and catch FileNotFoundException.
     *Read the file. Close it. */
    
    String str = "";      
    BufferedReader inputStream = null;
    try {
      inputStream = new BufferedReader(new FileReader(fileName));
      String l;
      while ((l = inputStream.readLine()) != null) {
          str = str +l+"\n"; 
      }
    } catch (FileNotFoundException e) {
        System.out.println("FileNotFoundException: " 
                        + e.getMessage());
        
    } catch (IOException e) {
         System.err.println("Caught IOException: " 
                             +  e.getMessage());
    } finally {
        if (inputStream != null) {
             System.out.println("Closing BufferReader");
             try { 
               inputStream.close(); 
             } catch (IOException e) { System.err.println("Caught closing Buffer IOException: " 
                             +  e.getMessage());			 
             }
         } 
         else {
             System.out.println("BufferReader not open");
         }
     return (str);
    }
}
  
  public void endSim () {
    aversionPlot.updateGraph();
    aversionPlot.takeSnapshot();
    
    aversionHistogram.step();
    aversionHistogram.takeSnapshot();
   
    
    String str;
    String fileName = "meanaversion_deathprob"+deathProb+
            "_advantage"+winnerAdvantage+".txt";
   
    str = ReadOldData (fileName );
     /* Open the file it using Print Writer.
      Apend new data to the already read data.
      Flush and close it. */ 
    PrintWriter outputStream = null;
    try {
      outputStream = new PrintWriter(new FileWriter(fileName));
      if (str == "" ) outputStream.println("maxAgentAge; globalMeanAversion;" +
              " globalMinAversion; globalMaxAversion");
      else outputStream.append(str);
      
      
      //outputStream.append(maxAgentAge+"; "+globalMeanAversion+"; "+globalMinAversion+
      //       "; "+globalMaxAversion);
      
      outputStream.println("finalPlot.plotPointWithErrorBars((double)"+
              maxAgentAge+","+globalMeanAversion+","+globalMinAversion+","+
              globalMaxAversion+",1);");
      
      outputStream.println("finalPlot.plotPoint((double)"+maxAgentAge+","+
              globalMeanAversion+", 0);");
      
      outputStream.flush();
      outputStream.close();
    }  catch (IOException e) {
         System.err.println("Caught IOException: " 
                             +  e.getMessage());
				 
    } finally {
         if (outputStream != null) {
             System.out.println("Closing PrintWriter");
             outputStream.close();

         } 
         else {
             System.out.println("PrintWriter not open");
         }
    };
    
    stop();
  
  }
  
  public void step() {
    
    for (int j=0; j<populations; j++) {
    
      agentList = agentPopulation [j];
      int aged =0;
      // Agents play games
      actualAgents = agentList.size();
      for (int i = 0; i < actualAgents; i++) {
        RiskAversionAgent a = (RiskAversionAgent)agentList.get(i);
        if (a.checkAge(maxAgentAge)) a.play(deathProb);
        else aged++;
      }

      int loosers=deathList.size()- aged;
      agentList.removeAll(deathList);
      deathList.clear();

      // Agents evolve
      if (agentList.isEmpty()) stop();
      evolution ();

      if (printOutput)
        System.out.println("Generation:"+getTickCount()+"Initial:"+actualAgents+
              " Loosers:"+loosers+" Aged:"+aged+
              " Births:"+birthList.size()+"\n"); 

      //agentList.removeAll(deathList);
      agentList.addAll(birthList);
      //deathList.clear();
      birthList.clear(); 
    
    if (agentList.isEmpty()) stop();
    meanAversion[j] = getMeanAversion();
    
    }
    
    if (getTickCount()>maxIterations) endSim();
    
    
    updateGlobalMeanMinMaxAversion();
    aversionHistogram.step();
    aversionPlot.plotPointWithErrorBars((double)getTickCount(), globalMeanAversion,
            globalMinAversion,globalMaxAversion, 1);
    aversionPlot.plotPoint((double)getTickCount(), globalMeanAversion, 0);
    aversionPlot.updateGraph();
  }

  public static void main(String[] args) throws IOException {
    uchicago.src.sim.engine.SimInit init =
            new uchicago.src.sim.engine.SimInit();
    RiskAversionModel model = new RiskAversionModel();
    init.loadModel(model, null, false);
  }

  // Methods for parameter manipulation ////////////////////////////////////
  
  public void setMaxAgentAge(int val) {
    maxAgentAge = val;
  }

  public int getMaxAgentAge() {
    return maxAgentAge;
  }

  public void setMaxIterations(int val) {
    maxIterations = val;
  }

  public int getMaxIterations() {
    return maxIterations;
  }
  
   public void setNumberOfAgents(int val) {
    numberOfAgents = val;
  }

  public int getNumberOfAgents() {
    return numberOfAgents;
  }
  
  public void setPopulations(int val) {
    if (val>0) populations = val;
    else populations=1;
  }

  public int getPopulations() {
    return populations;
  }
/* 
  public void setActualAgents(int val) {
    actualAgents = val;
  }
*/
    public int getActualAgents() {
    return actualAgents;
  }

    public void setDeathProb(double val) {
    deathProb = val;
  }
       
  public double getDeathProb() {
    return deathProb;
  }

    public void setMutationRange(double val) {
    mutationRange = val;
  }
       
  public double getMutationRange() {
    return mutationRange;
  }

  public void setWinnerAdvantage(double val) {
    winnerAdvantage = val;
  }
       
  public double getWinnerAdvantage() {
    return winnerAdvantage;
  }

  public void setBaseReproductionProb(double val) {
    baseReproductionProb = val;
  }
  
  public double getBaseReproductionProb() {
    return baseReproductionProb;
  }

  
  public void setAutoReproductionProb(boolean val) {
    autoReproductionProb = val;
  }
    
  public boolean getAutoReproductionProb() {
    return autoReproductionProb;
  }

   public void setFixedPopulation(boolean val) {
    fixedPopulation = val;
  }
    
  public boolean getFixedPopulation() {
    return fixedPopulation;
  }

  public void setPrintOutput(boolean val) {
    printOutput = val;
  }
    
  public boolean getPrintOutput() {
    return printOutput;
  }

    public void setSexualReproduction(boolean val) {
    sexualReproduction = val;
  }
    
  public boolean getSexualReproduction() {
    return sexualReproduction;
  }
}  

